module Types(Type(..),TypeList,TVar,vars, vars', tcons, close, errType, 
             restype,argtype,
             Sub,idsub,single,bar,without,apply,comp,
             unify,unifySet,unif,Unif(..),
             tmatch,Env,dom,ran,app,Lookup(..),
             appid,assign,unknowns,fcomp,fcom,
             namesupply,generalize,inst,nbound,tbs) where

import List

type TypeList = [Type]

-- Type variables

type TVar = String

-- Type expressions

data Type = TCon String [Type] | Tid TVar | TAbs [String] Type 
			deriving (Show, Eq)

vars t = nub (vars' [] t []) 

vars' zs (TCon x ts) ys = foldr (vars' zs) ys ts
vars' zs (Tid x) ys = if x `elem` zs then ys else x:ys
vars' zs (TAbs xs t) ys = vars' (xs++zs) t ys

tcons (TCon x ts) = x :(concat (map tcons ts))
tcons (Tid _) = []
tcons (TAbs _ t) = tcons t

close t = if null vt then t else TAbs nt (apply (appid (zip vt (map Tid nt))) t)
          where vt = vars t
                nt = take (length vt) namesupply

errType = TAbs ["a"] (Tid "a")

restype (TCon "Arrow" [_,t]) = t
restype _ = error "cannot take result type of non-function"

argtype (TCon "Arrow" [t,_]) = t
argtype _ = error "cannot take argument type of non-function"

-- Substitutions 

type Sub = TVar -> Type

idsub :: Sub
idsub = Tid

single :: TVar -> Type -> Sub
single x t y = if x==y then t else Tid y

bar :: TVar -> Sub -> Sub 
bar x s y = if x==y then Tid x else s y

without :: [TVar] -> Sub -> Sub
without xs s = foldr bar s xs

-- we assume hygienic naming
 
apply :: Sub -> Type -> Type
apply s (TCon name ts) = TCon name (map (apply s) ts)
apply s (Tid name) = s name
apply s (TAbs xs t) = TAbs xs (apply (without xs s) t)

comp :: Sub -> Sub -> Sub
comp s t = apply s . t 

unify :: Type -> Type -> Unif
unify x y = unif [(x,y)] idsub

unifySet :: [(Type,Type)] -> Unif
unifySet ps = unif ps idsub

pap :: Sub -> (Type,Type) -> (Type,Type)
pap s (t,u) =(apply s t, apply s u)

-- only first order unification

data Unif = Success Sub | Mismatch Type Type | Occurs TVar Type

unif :: [(Type,Type)] -> Sub -> Unif
unif [] sub = Success sub
unif ((TCon x ts,TCon y ss):ps) sub | x==y && length ts == length ss = unif (zip ts ss ++ ps) sub
                                    | otherwise = Mismatch (TCon x ts) (TCon y ss)
unif ((Tid x,Tid y):ps) sub | x==y        = unif ps sub
                            | otherwise   = unif (map (pap d) ps) (d `comp` sub)
                                            where d = single x (Tid y)
unif ((Tid x,t):ps) sub | x `elem` vars t = Occurs x t
                        | otherwise       = unif (map (pap d) ps) (d `comp` sub)
                                            where d = single x t
unif ((t,Tid x):ps) sub | x `elem` vars t = Occurs x t
                        | otherwise       = unif (map (pap d) ps) (d `comp` sub)
                                            where d = single x t
unif ((s,t):ps) sub = if (s==t) then unif ps sub
                      else error "internal error in unification" 

-- type matching

tmatch :: Type -> Type -> [Sub]
tmatch x y = txmatch idsub (x,y) 

textend :: Sub -> (String,Type) -> [Sub]
textend s (v,x)  
   | y == x     = [s]
   | y == Tid v = [comp (single v x) s]
   | otherwise  = []
   where y = s v

txmatch :: Sub -> (Type,Type) -> [Sub]
txmatch s (Tid v,x) = textend s (v,x)
txmatch s (TCon f xs,TCon g ys) = if f==g
                                  then txmatchlist s (zip xs ys) 
                                  else []
txmatch s (_ , _) = []

txmatchlist :: Sub -> [(Type,Type)] -> [Sub]
txmatchlist s [] = [s]
txmatchlist s (xy:xys) = concat [ txmatchlist t xys | t <- txmatch s xy]

-- Environments

type Env = [(String, Type)]

data Lookup = AType Type | None String

dom :: Env -> [TVar]
dom = nub . map fst

ran :: Env -> [Type]
ran e = map (appid e) (dom e)

app :: Env -> String -> Lookup
app env x = if null ts then None x
            else AType (head ts)
            where ts = [ t | (y,t) <- env, x==y]

appid :: Env -> String -> Type
appid env x = if null ts then Tid x else head ts
              where ts = [ t | (y,t) <- env, x==y]

assign :: Env -> TVar -> Type -> Env
assign env x t = (x,t):env

unknowns :: Env -> [TVar]
unknowns = nub . concat . map vars . ran

fcomp :: (Type -> Type) -> Env -> Env
fcomp f env = [(x,f t) | (x,t) <- env]

fcom :: Sub -> Env -> Env
fcom s = fcomp (apply s) 

join :: Env -> Env -> Env
join env1 env2 = [(x,t) | (x,t) <- env1, not (x `elem` dom env2)] ++ env2

restrict :: Env -> [String] -> Env
restrict env xs = [(x,t) | (x,t) <- env, x `elem` xs ]

-- name supply

namesupply = tail (concat (iterate next [[]]))
             where next xs = [a:x | a<-['a'..'z'], x<-xs]

generalize bar genv 
 = foldr step [] genv 
   where step (x,t) env
          = assign env x  (tbs nxs (apply (appid renaming) t)) 
            where nxs = take (length vtbar) (namesupply \\ bar)
                  vtbar = vars t \\ bar 
                  renaming  = zip vtbar (map Tid nxs) 
                

inst [] t = t 
inst fresh (TAbs xs t) = apply (appid (zip xs (map Tid fresh))) t

nbound (TAbs xs t) = length xs
nbound t = 0

tbs [] t = t
tbs xs t = TAbs xs t
